@@ -1,5 +1,9 @@
 // SPDX-License-Identifier: BSL-1.0
 
+// _FORTIFY_SOURCE does not work without optimiziation, but we are going to disable optimization later in this file.
+// So just undefine it, to avoid build failures.
+#undef _FORTIFY_SOURCE
+
 #if __has_include(<aio.h>)
 #include <aio.h>
 #endif
@@ -19,10 +23,18 @@
 #include <sys/stropts.h>
 #endif
 
+#if defined(TEST_ENABLE_FERAISEEXCEPT)
+#include <fenv.h>
+#endif
+
 #define CATCH_CONFIG_EXTERNAL_INTERFACES
 #define CATCH_CONFIG_NO_POSIX_SIGNALS
 #ifndef BUNDLED_CATCH2
+#ifdef CATCH3
+#include "catch2/catch_all.hpp"
+#else
 #include "catch2/catch.hpp"
+#endif
 #else
 #include "catch.hpp"
 #endif
@@ -34,12 +46,36 @@
 
 #include "PosixSignalManager.h"
 
-struct SaneStateListener : Catch::TestEventListenerBase {
-    SaneStateListener(Catch::ReporterConfig const& _config) : Catch::TestEventListenerBase(_config) {
+// This test is by necessity made using undefined behavior (after all it needs to test what happens in situations that
+// are for good reason undefined behavior).
+// To make this a bit more likely to work, just disable all optimizations if possible.
+
+#ifdef __clang__
+#pragma clang optimize off
+#else
+#ifdef __GNUC__
+#pragma GCC optimize ("-O0")
+#endif
+#endif
+
+
+#ifdef CATCH3
+using ListenerBase = Catch::EventListenerBase;
+#else
+using ListenerBase = Catch::TestEventListenerBase;
+#endif
+
+struct SaneStateListener : ListenerBase {
+
+#if CATCH3
+    SaneStateListener(Catch::IConfig const* config) : ListenerBase(config) {
+#else
+    SaneStateListener(Catch::ReporterConfig const& _config) : ListenerBase(_config) {
+#endif
         mainPid = getpid();
     }
 
-    using TestEventListenerBase::TestEventListenerBase; // inherit constructor
+    using ListenerBase::ListenerBase; // inherit constructor
 
     virtual void testCaseStarting(Catch::TestCaseInfo const& testInfo) override {
         (void)testInfo;
@@ -56,12 +92,20 @@ struct SaneStateListener : Catch::TestEventListenerBase {
     virtual void testCaseEnded(Catch::TestCaseStats const& testCaseStats) override {
         if (getpid() != mainPid) {
             puts("FATAL: child process did escape into test runner. Test:");
+#if CATCH3
+            puts(testCaseStats.testInfo->name.c_str());
+#else
             puts(testCaseStats.testInfo.name.c_str());
+#endif
             abort();
         }
         if (PosixSignalManager::isCreated()) {
             puts("FATAL: PosixSignalManager was created in test runner process. Test:");
+#if CATCH3
+            puts(testCaseStats.testInfo->name.c_str());
+#else
             puts(testCaseStats.testInfo.name.c_str());
+#endif
             abort();
         }
     }
@@ -70,6 +114,16 @@ struct SaneStateListener : Catch::TestEventListenerBase {
 };
 CATCH_REGISTER_LISTENER(SaneStateListener)
 
+void resetSignalsCatch3() {
+#if CATCH3
+   // There doesn't seem to be a way to disable signal catching that works with
+   // distro precompiled catch library. Workaround by resetting manually.
+   for (int signo = 0; signo < 128; signo++) {
+        signal(signo, SIG_DFL);
+    }
+#endif
+}
+
 #ifdef __linux__
 bool isLinux() {
     utsname uts;
@@ -118,6 +172,11 @@ void cause_sigsegv() {
 int zero = 0;
 
 void cause_sigfpe() {
+#if defined(TEST_ENABLE_FERAISEEXCEPT)
+    feenableexcept(FE_DIVBYZERO);
+    feraiseexcept(FE_DIVBYZERO);
+#endif
+
     printf("%d\n", 3 / zero);
 }
 
@@ -226,6 +285,11 @@ struct SharedPageAlloc {
     INFO("info=" << std::hex << info);     \
     REQUIRE(r != -1)
 
+#define PREDICATE_SIGNALED_WITH(signo)  \
+    ((r == pid)                         \
+     && (WIFSIGNALED(info))             \
+     && (WTERMSIG(info) == signo))
+
 #define WAS_SIGNALED_WITH(signo)        \
     CHECK(r == pid);                    \
     CHECK(WIFSIGNALED(info));           \
@@ -252,6 +316,11 @@ struct SharedPageAlloc {
     int r = waitid(P_ALL, 0, &info, WEXITED | WSTOPPED | WCONTINUED);   \
     REQUIRE(r == 0)
 
+#define PREDICATE_SIGNALED_WITH(signo)  \
+    ((info.si_pid == pid)               \
+     && (info.si_code == CLD_KILLED)    \
+     && (info.si_status == signo))
+
 #define WAS_SIGNALED_WITH(signo)        \
     CHECK(info.si_pid == pid);          \
     CHECK(info.si_code == CLD_KILLED);  \
@@ -289,6 +358,7 @@ struct SharedPageAlloc {
 #endif
 
 TEST_CASE( "signal classification" ) {
+    resetSignalsCatch3();
     int signo = GENERATE(range(1, NUM_SIGNALS));
     CAPTURE(signo);
     if (signo == SIGKILL || signo == SIGSTOP || signo == SIGTSTP || signo == SIGTTIN || signo == SIGTTOU) {
@@ -327,6 +397,7 @@ TEST_CASE( "signal classification" ) {
 }
 
 TEST_CASE( "baseline sigsegv" ) {
+    resetSignalsCatch3();
     pid_t pid = fork();
     REQUIRE(pid != -1);
     if (pid) {
@@ -339,6 +410,7 @@ TEST_CASE( "baseline sigsegv" ) {
 }
 
 TEST_CASE( "reraise sigsegv" ) {
+    resetSignalsCatch3();
     SharedPageAlloc sharedPageAlloc;
     pid_t pid = fork();
     REQUIRE(pid != -1);
@@ -358,6 +430,7 @@ TEST_CASE( "reraise sigsegv" ) {
 }
 
 TEST_CASE( "reraise sigsegv in fork" ) {
+    resetSignalsCatch3();
     int follow = GENERATE(0, 1, 2);
     CAPTURE(follow);
     SharedPageAlloc sharedPageAlloc;
@@ -403,6 +476,7 @@ TEST_CASE( "reraise sigsegv in fork" ) {
 }
 
 TEST_CASE( "reraise 'raised' sigsegv" ) {
+    resetSignalsCatch3();
     SharedPageAlloc sharedPageAlloc;
     pid_t pid = fork();
     REQUIRE(pid != -1);
@@ -430,6 +504,7 @@ TEST_CASE( "reraise 'raised' sigsegv" ) {
 }
 
 TEST_CASE( "reraise 'killed' sigsegv" ) {
+    resetSignalsCatch3();
     SharedPageAlloc sharedPageAlloc;
     pid_t pid = fork();
     REQUIRE(pid != -1);
@@ -454,6 +529,7 @@ TEST_CASE( "reraise 'killed' sigsegv" ) {
 #if defined(SIGRTMIN)
 // ^^^ sigqueue depends on realtime signals
 TEST_CASE( "reraise 'queued' sigsegv" ) {
+    resetSignalsCatch3();
     SharedPageAlloc sharedPageAlloc;
     pid_t pid = fork();
     REQUIRE(pid != -1);
@@ -478,6 +554,7 @@ TEST_CASE( "reraise 'queued' sigsegv" ) {
 
 #ifdef __linux__
 TEST_CASE( "reraise 'io' sigsegv" ) {
+    resetSignalsCatch3();
     bool linux_pre_4_14 = false;
     if (isLinux() && utsRelease() < QVector<int>{4,14}) {
         WARN("\"reraise 'io' sigsegv\" test uses degraded handling on linux kernel < 4.14");
@@ -532,6 +609,7 @@ bad:
 #if defined(SIGRTMIN)
 // ^^^ timer_create depends on realtime signals
 TEST_CASE( "reraise 'timer' sigsegv" ) {
+    resetSignalsCatch3();
     SharedPageAlloc sharedPageAlloc;
     pid_t pid = fork();
     REQUIRE(pid != -1);
@@ -573,6 +651,7 @@ TEST_CASE( "reraise 'timer' sigsegv" ) {
 #if defined(__FreeBSD__)
 // FreeBSD doesn't implement mq_open (it only returns ENOSYS)
 TEST_CASE( "mqueue not implemented" ) {
+    resetSignalsCatch3();
     const char *name = "/PosixSignalManager-test";
     mq_unlink(name);
     errno = 0;
@@ -588,6 +667,7 @@ TEST_CASE( "mqueue not implemented" ) {
 
 #if HAVE_MQ
 TEST_CASE( "reraise 'mq' sigsegv" ) {
+    resetSignalsCatch3();
     SharedPageAlloc sharedPageAlloc;
     pid_t pid = fork();
     REQUIRE(pid != -1);
@@ -630,6 +710,7 @@ TEST_CASE( "reraise 'mq' sigsegv" ) {
 
 #if __has_include(<aio.h>)
 TEST_CASE( "reraise 'aio' sigsegv" ) {
+    resetSignalsCatch3();
     SharedPageAlloc sharedPageAlloc;
     pid_t pid = fork();
     REQUIRE(pid != -1);
@@ -678,6 +759,7 @@ TEST_CASE( "reraise 'aio' sigsegv" ) {
 
 #ifndef NO_SIGBUS
 TEST_CASE( "baseline sigbus" ) {
+    resetSignalsCatch3();
     checkdeps_sigbus();
     pid_t pid = fork();
     REQUIRE(pid != -1);
@@ -693,6 +775,7 @@ TEST_CASE( "baseline sigbus" ) {
 
 #ifndef NO_SIGBUS
 TEST_CASE( "reraise sigbus" ) {
+    resetSignalsCatch3();
     checkdeps_sigbus();
     SharedPageAlloc sharedPageAlloc;
     pid_t pid = fork();
@@ -713,6 +796,7 @@ TEST_CASE( "reraise sigbus" ) {
 #endif
 
 TEST_CASE( "reraise 'killed' sigbus" ) {
+    resetSignalsCatch3();
     SharedPageAlloc sharedPageAlloc;
     pid_t pid = fork();
     REQUIRE(pid != -1);
@@ -739,6 +823,7 @@ TEST_CASE( "reraise 'killed' sigbus" ) {
 #if !defined(__FreeBSD__) && !defined(__OpenBSD__) && !defined(__APPLE__) && !defined(__NetBSD__)
 // ^^^ on {freebsd,openbsd,apple,netbsd} sigio is ignored by default, so reraise will not kill the child
 TEST_CASE( "reraise sigio" ) {
+    resetSignalsCatch3();
     REQUIRE(PosixSignalManager::classifySignal(SIGIO) > 0);
     SharedPageAlloc sharedPageAlloc;
     pid_t pid = fork();
@@ -791,6 +876,7 @@ bad:
 }
 #else
 TEST_CASE( "check sigio" ) {
+    resetSignalsCatch3();
     REQUIRE(PosixSignalManager::classifySignal(SIGIO) == 0);
     SharedPageAlloc sharedPageAlloc;
     pid_t pid = fork();
@@ -844,6 +930,7 @@ bad:
 #endif
 
 TEST_CASE( "reraise 'killed' sigio" ) {
+    resetSignalsCatch3();
     SharedPageAlloc sharedPageAlloc;
     pid_t pid = fork();
     REQUIRE(pid != -1);
@@ -871,6 +958,7 @@ TEST_CASE( "reraise 'killed' sigio" ) {
 
 #ifndef NO_SIGILL
 TEST_CASE( "baseline sigill" ) {
+    resetSignalsCatch3();
     pid_t pid = fork();
     REQUIRE(pid != -1);
     if (pid) {
@@ -885,6 +973,7 @@ TEST_CASE( "baseline sigill" ) {
 
 #ifndef NO_SIGILL
 TEST_CASE( "reraise sigill" ) {
+    resetSignalsCatch3();
     SharedPageAlloc sharedPageAlloc;
     pid_t pid = fork();
     REQUIRE(pid != -1);
@@ -904,6 +993,7 @@ TEST_CASE( "reraise sigill" ) {
 #endif
 
 TEST_CASE( "reraise 'killed' sigill" ) {
+    resetSignalsCatch3();
     SharedPageAlloc sharedPageAlloc;
     pid_t pid = fork();
     REQUIRE(pid != -1);
@@ -925,7 +1015,14 @@ TEST_CASE( "reraise 'killed' sigill" ) {
     }
 }
 
+// triggering SIGFPE portably is hard and some architectures like aarch64 even seem to allow implementation to
+// not implement it at all or even don't have a concept like it at all (risc v)
+// So require it only for known good architectures and only do opportunistic testing otherwise.
+
+#if defined(__amd64__) || defined(__i386__) || defined(__mips__)
+
 TEST_CASE( "baseline sigfpe" ) {
+    resetSignalsCatch3();
     pid_t pid = fork();
     REQUIRE(pid != -1);
     if (pid) {
@@ -937,7 +1034,34 @@ TEST_CASE( "baseline sigfpe" ) {
     }
 }
 
+#endif
+
 TEST_CASE( "reraise sigfpe" ) {
+    resetSignalsCatch3();
+
+    // Check if triggering SIGFPE is supported on this platform
+    bool supported = false;
+    {
+        pid_t pid = fork();
+        REQUIRE(pid != -1);
+        if (pid) {
+            WAIT_CHILD;
+            supported = PREDICATE_SIGNALED_WITH(SIGFPE);
+        } else {
+            cause_sigfpe();
+            _exit(99);
+        }
+    }
+
+    if (!supported) {
+        WARN("Skipping 'reraise sigfpe' test, because triggering SIGFPE failed");
+        return;
+    } else {
+#if defined(__riscv)
+        FAIL("Archtecture skip-list contains architecture where SIGFPE could be triggered");
+#endif
+    }
+
     SharedPageAlloc sharedPageAlloc;
     pid_t pid = fork();
     REQUIRE(pid != -1);
@@ -956,6 +1080,7 @@ TEST_CASE( "reraise sigfpe" ) {
 }
 
 TEST_CASE( "reraise 'killed' sigfpe" ) {
+    resetSignalsCatch3();
     SharedPageAlloc sharedPageAlloc;
     pid_t pid = fork();
     REQUIRE(pid != -1);
@@ -979,6 +1104,7 @@ TEST_CASE( "reraise 'killed' sigfpe" ) {
 
 #ifndef NO_SIGTRAP
 TEST_CASE( "baseline sigtrap" ) {
+    resetSignalsCatch3();
     pid_t pid = fork();
     REQUIRE(pid != -1);
     if (pid) {
@@ -993,6 +1119,7 @@ TEST_CASE( "baseline sigtrap" ) {
 
 #ifndef NO_SIGTRAP
 TEST_CASE( "reraise sigtrap" ) {
+    resetSignalsCatch3();
     SharedPageAlloc sharedPageAlloc;
     pid_t pid = fork();
     REQUIRE(pid != -1);
@@ -1012,6 +1139,7 @@ TEST_CASE( "reraise sigtrap" ) {
 #endif
 
 TEST_CASE( "reraise 'killed' sigtrap" ) {
+    resetSignalsCatch3();
     SharedPageAlloc sharedPageAlloc;
     pid_t pid = fork();
     REQUIRE(pid != -1);
@@ -1037,6 +1165,7 @@ TEST_CASE( "reraise 'killed' sigtrap" ) {
     && !defined(__NetBSD__) && !defined(__sun)
 // ^^^ various bsds do not have a way to change the signal for O_ASYNC
 TEST_CASE( "reraise 'io' sigrt" ) {
+    resetSignalsCatch3();
     SharedPageAlloc sharedPageAlloc;
     pid_t pid = fork();
     REQUIRE(pid != -1);
@@ -1079,6 +1208,7 @@ bad:
 #endif
 
 TEST_CASE( "term handler (sighup)" ) {
+    resetSignalsCatch3();
     SharedPageAlloc sharedPageAlloc;
     pid_t pid = fork();
     REQUIRE(pid != -1);
@@ -1103,6 +1233,7 @@ TEST_CASE( "term handler (sighup)" ) {
 }
 
 TEST_CASE( "term handler (sighup) in fork" ) {
+    resetSignalsCatch3();
     int follow = GENERATE(0, 1, 2);
     CAPTURE(follow);
     SharedPageAlloc sharedPageAlloc;
@@ -1148,6 +1279,7 @@ TEST_CASE( "term handler (sighup) in fork" ) {
 
 #if defined(SIGRTMIN)
 TEST_CASE( "term handler (sigrt)" ) {
+    resetSignalsCatch3();
     SharedPageAlloc sharedPageAlloc;
     pid_t pid = fork();
     REQUIRE(pid != -1);
@@ -1173,6 +1305,7 @@ TEST_CASE( "term handler (sigrt)" ) {
 #endif
 
 TEST_CASE( "ignored term handler (sighup)" ) {
+    resetSignalsCatch3();
     SharedPageAlloc sharedPageAlloc;
     pid_t pid = fork();
     REQUIRE(pid != -1);
@@ -1191,6 +1324,7 @@ TEST_CASE( "ignored term handler (sighup)" ) {
 }
 
 TEST_CASE( "removed term handler (sighup)" ) {
+    resetSignalsCatch3();
     SharedPageAlloc sharedPageAlloc;
     pid_t pid = fork();
     REQUIRE(pid != -1);
@@ -1210,6 +1344,7 @@ TEST_CASE( "removed term handler (sighup)" ) {
 }
 
 TEST_CASE( "crash handler (sigsegv)" ) {
+    resetSignalsCatch3();
     SharedPageAlloc sharedPageAlloc;
     pid_t pid = fork();
     REQUIRE(pid != -1);
@@ -1229,6 +1364,7 @@ TEST_CASE( "crash handler (sigsegv)" ) {
 }
 
 TEST_CASE( "crash handler (sigsegv) in fork" ) {
+    resetSignalsCatch3();
     int follow = GENERATE(0, 1, 2);
     CAPTURE(follow);
     SharedPageAlloc sharedPageAlloc;
@@ -1274,6 +1410,7 @@ TEST_CASE( "crash handler (sigsegv) in fork" ) {
 }
 
 TEST_CASE( "removed crash handler (sigsegv)" ) {
+    resetSignalsCatch3();
     SharedPageAlloc sharedPageAlloc;
     pid_t pid = fork();
     REQUIRE(pid != -1);
@@ -1293,6 +1430,7 @@ TEST_CASE( "removed crash handler (sigsegv)" ) {
 }
 
 TEST_CASE( "notify (sighup)" ) {
+    resetSignalsCatch3();
     SharedPageAlloc sharedPageAlloc;
     pid_t pid = fork();
     REQUIRE(pid != -1);
@@ -1331,6 +1469,7 @@ TEST_CASE( "notify (sighup)" ) {
 }
 
 TEST_CASE( "removed notify (sighup)" ) {
+    resetSignalsCatch3();
     SharedPageAlloc sharedPageAlloc;
     pid_t pid = fork();
     REQUIRE(pid != -1);
@@ -1367,6 +1506,7 @@ TEST_CASE( "removed notify (sighup)" ) {
 }
 
 TEST_CASE( "removed notify (sigchld)" ) {
+    resetSignalsCatch3();
     SharedPageAlloc sharedPageAlloc;
     pid_t pid = fork();
     REQUIRE(pid != -1);
@@ -1403,6 +1543,7 @@ TEST_CASE( "removed notify (sigchld)" ) {
 }
 
 TEST_CASE( "notify (sighup) in fork" ) {
+    resetSignalsCatch3();
     int follow = GENERATE(0, 1, 2);
     CAPTURE(follow);
     SharedPageAlloc sharedPageAlloc;
@@ -1461,6 +1602,7 @@ TEST_CASE( "notify (sighup) in fork" ) {
 }
 
 TEST_CASE( "reraise TSTP/TTIN/TTOU" ) {
+    resetSignalsCatch3();
     int signo = GENERATE(SIGTSTP, SIGTTIN, SIGTTOU);
     CAPTURE(signo);
     SharedPageAlloc sharedPageAlloc;
@@ -1499,6 +1641,7 @@ TEST_CASE( "reraise TSTP/TTIN/TTOU" ) {
 }
 
 TEST_CASE( "brute force 'killed' reraise" ) {
+    resetSignalsCatch3();
     int signo = GENERATE(range(1, NUM_SIGNALS));
     CAPTURE(signo);
     if (signo == SIGKILL || signo == SIGSTOP || signo == SIGTSTP || signo == SIGTTIN || signo == SIGTTOU) {
